import pathlib
from typing import Optional, IO, TextIO, BinaryIO, Any

from rdflib import Graph, URIRef
from rdflib.parser import InputSource
from rdflib.term import Node


def copy_triples(src: Graph, dest: Graph):
    """Add all triples in `src` to `dest`."""
    for triple in src:
        dest.add(triple)


def update_node(node: Node, old_uri: URIRef, new_uri: URIRef) -> Node:
    """If `node` equals the `old_uri` (with an optional fragment identifier),
    replace `old_uri` with `new_uri` and return a new URIRef object. Otherwise,
    return the original `node`."""
    if node == old_uri or str(node).startswith(old_uri + '#'):
        return URIRef(str(node).replace(old_uri, new_uri))
    else:
        return node


def new_triple(old_uri: URIRef, new_uri: URIRef, s: Node, p: Node, o: Node) -> tuple[Node, Node, Node]:
    """Update instances of `old_uri` to `new_uri` in `s`, `p`, and `o`,
    and return the updated triple."""
    new_s = update_node(s, old_uri, new_uri)
    new_p = update_node(p, old_uri, new_uri)
    new_o = update_node(o, old_uri, new_uri)
    return new_s, new_p, new_o


class TrackChangesGraph(Graph):
    """An RDF graph that tracks inserts and deletes."""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.original = Graph()
        """Original graph"""

    def parse(
        self,
        source: Optional[
            IO[bytes] | TextIO | InputSource | str | bytes | pathlib.PurePath
        ] = None,
        publicID: Optional[str] = None,  # noqa: N803
        format: Optional[str] = None,
        location: Optional[str] = None,
        file: Optional[BinaryIO | TextIO] = None,
        data: Optional[str | bytes] = None,
        **args: Any,
    ) -> 'TrackChangesGraph':
        """Parses the graph normally, and then saves a copy of the original."""
        super().parse(source, publicID, format, location, file, data, **args)
        self.original = Graph()
        copy_triples(self, self.original)
        return self

    def change_uri(self, old_uri: URIRef, new_uri: URIRef):
        """Change occurrences of ``old_uri`` to ``new_uri`` in this graph.
        This includes URIRefs that contain a fragment identifier following
        the ``old_uri``.

        This object is updated in place."""
        for s, p, o in self:
            new_s, new_p, new_o = new_triple(old_uri, new_uri, s, p, o)
            self.remove((s, p, o))
            self.add((new_s, new_p, new_o))

    @property
    def inserts(self) -> Graph:
        """Graph containing triples that have been added"""
        return self - self.original

    @property
    def deletes(self) -> Graph:
        """Graph containing triples that have been removed"""
        return self.original - self

    @property
    def has_changes(self) -> bool:
        """Whether this graph has been changed"""
        return len(self.inserts) > 0 or len(self.deletes) > 0

    def apply_changes(self):
        """Replace the original graph with the current graph. Immediately
        after calling this method, `has_changes()` will return `False`."""
        graph = Graph()
        copy_triples(self, graph)
        self.original = graph



from copy import deepcopy, copy

from rdflib import Literal

from plastron.rdfmapping.resources import RDFResource


def test_add_remove_values():
    resource = RDFResource(label=Literal('foo'))
    assert len(resource.label) == 1
    resource.label.add(Literal('bar'))
    assert len(resource.label) == 2
    resource.label.clear()
    assert len(resource.label) == 0


def test_same_obj_same_prop():
    resource = RDFResource()
    # multiple accesses of the property for the same resource object
    # returns the same property object; that is, as long as `a is b`
    # holds, so should `a.prop is b.prop`
    assert resource.label is resource.label
    other = resource
    assert other.label is resource.label
    # shallow copies also retain the same property objects; this is
    # due to the semantics of the `copy` operation, and is unlikely
    # to be very useful in production code; where a copy of a resource
    # is needed, `deepcopy` should be used instead
    shallow_copy = copy(resource)
    assert shallow_copy.label is resource.label


def test_different_obj_different_prop():
    a = RDFResource()
    b = RDFResource()
    assert a.label is not b.label
    c = deepcopy(a)
    assert a.label is not c.label


def test_property_preserves_value():
    resource = RDFResource()
    value = Literal('foo')
    resource.label = value
    assert resource.label.value is value



from collections import defaultdict
from copy import deepcopy, copy
from typing import Optional, Any, Type, TypeVar, Callable
from uuid import uuid4

from rdflib import Graph, URIRef

from plastron.rdfmapping.descriptors import ObjectProperty, Property, DataProperty, OBJECT_CLASSES
from plastron.rdfmapping.graph import TrackChangesGraph, copy_triples
from plastron.rdfmapping.properties import RDFProperty
from plastron.rdfmapping.validation import ValidationResultsDict


def is_iterable(value: Any) -> bool:
    """Returns `True` if `value` is iterable, but not a string. While strings
    in Python are technically iterable, there are many cases where we would
    prefer to not treat them as such."""
    try:
        iter(value)
    except TypeError:
        # this is non-iterable
        return False
    else:
        # special case for strings; they are technically iterable,
        # but in all but the most specialized cases we want to treat
        # them as single values
        return not isinstance(value, str)


class RDFResourceBase:
    """Base class for RDF description classes."""
    rdf_property_names: set = set()
    default_values: dict[Any, set] = defaultdict(set)
    validators: list[Callable[['RDFResourceBase'], bool]] = []

    def __init_subclass__(cls, **kwargs):
        # make new copies of the class variables for the subclasses
        # at this point, the Property descriptors' __set_name__ methods
        # have already run, so we start with the names of this class's
        # own Property descriptors
        cls.rdf_property_names = {k for k, v in cls.__dict__.items() if isinstance(v, Property)}
        cls.default_values = defaultdict(set)
        cls.validators = []
        base_classes = list(filter(lambda c: issubclass(c, RDFResourceBase), cls.__mro__))
        for base_class in base_classes:
            cls.rdf_property_names |= copy(base_class.rdf_property_names)
            # default_values and validators are modified by decorators, which
            # run after the __init_subclass__ method
            cls.default_values.update(deepcopy(base_class.default_values))
            cls.validators.extend(deepcopy(base_class.validators))
        # build the lookup table of class name to class
        OBJECT_CLASSES[cls.__name__] = cls

    def __init__(self, uri: URIRef | str = None, graph: Graph = None, **kwargs):
        if uri is not None:
            self._uri = URIRef(uri)
        else:
            # use a unique UUID URI for resources that do not (yet) have a URI subject
            self._uri = URIRef(uuid4().urn)

        if graph:
            if isinstance(graph, TrackChangesGraph):
                self._graph = graph
            else:
                self._graph = TrackChangesGraph()
                copy_triples(graph, self._graph)
                copy_triples(graph, self._graph.original)
        else:
            self._graph = TrackChangesGraph()
        self.add_properties(**self.default_values)
        self.add_properties(**kwargs)

    def get_fragment_resource(
            self,
            object_class: Type['RDFResourceBase'],
            fragment_id: Optional[str] = None,
    ) -> 'RDFResourceBase':
        """
        Embedded (i.e., "fragment") resources share a graph with their parent resource. They
        are essentially a different "filter" through which to view the same graph.

        :param fragment_id: fragment identifier to append to the parent resource's URI. Defaults to a new UUIDv4 string.
        :param object_class:
        :return:
        """
        if fragment_id is None:
            fragment_id = str(uuid4())
        uri = URIRef(self.uri + '#' + fragment_id)
        fragment = object_class(uri=uri, graph=self._graph)
        return fragment

    def _update_properties(self, properties: dict[str, Any], clear_existing=False):
        for name, value in properties.items():
            if name not in self.rdf_property_names:
                raise ValueError(f'Unknown property name: {name}')

            prop = getattr(self, name)
            if clear_existing:
                prop.clear()
            if is_iterable(value):
                # value is iterable, add each value separately
                for v in value:
                    prop.add(v)
            else:
                # this is non-iterable, set a single value
                prop.add(value)

    def set_properties(self, **kwargs):
        """Set the properties on this object. Clears any existing properties."""
        self._update_properties(properties=kwargs, clear_existing=True)

    def add_properties(self, **kwargs):
        """Add the given properties to this object. Does *not* clear existing properties."""
        self._update_properties(properties=kwargs, clear_existing=False)

    @property
    def graph(self) -> TrackChangesGraph:
        return self._graph

    @property
    def has_changes(self) -> bool:
        return self._graph.has_changes

    def apply_changes(self):
        self._graph.apply_changes()

    @property
    def uri(self) -> URIRef:
        """The primary subject for this RDF resource."""
        return self._uri

    @uri.setter
    def uri(self, new_uri: URIRef):
        self._graph.change_uri(self._uri, new_uri)
        self._uri = new_uri

    def rdf_properties(self) -> list[RDFProperty]:
        return [getattr(self, attr_name) for attr_name in self.rdf_property_names]

    @property
    def is_valid(self) -> bool:
        if not all(p.is_valid for p in self.rdf_properties()):
            return False
        if not all(test(self) for test in self.validators):
            return False
        return True

    def validate(self) -> ValidationResultsDict:
        results = ValidationResultsDict({name: getattr(self, name).is_valid for name in self.rdf_property_names})
        for test in self.validators:
            results['_' + test.__name__] = test(self)
        return results

    def redescribe(self, model: Type['RDFResourceType']) -> 'RDFResourceType':
        return model(uri=self.uri, graph=self.graph)


class RDFResource(RDFResourceBase):
    rdf_type = ObjectProperty(URIRef('http://www.w3.org/1999/02/22-rdf-syntax-ns#type'), repeatable=True)
    label = DataProperty(URIRef('http://www.w3.org/2000/01/rdf-schema#label'), repeatable=True)


class ResourceError(Exception):
    pass


RDFResourceType = TypeVar('RDFResourceType', bound=RDFResourceBase)



from typing import Any, Callable, Container

from rdflib import URIRef, Literal

from plastron.rdfmapping.embed import EmbeddedObject
from plastron.rdfmapping.properties import RDFDataProperty, RDFObjectProperty, RDFProperty

OBJECT_CLASSES = {}


class Property:
    def __init__(
            self,
            predicate: URIRef,
            required: bool = False,
            repeatable: bool = False,
            values_from: Container = None,
            validate: Callable[[Any], bool] = None,
    ):
        self.predicate = predicate
        self.required = required
        self.repeatable = repeatable
        self.values_from = values_from
        self.validate = validate

    def __set_name__(self, owner, name):
        self.name = name
        self.private_name = f'_{name}'

    def __get__(self, instance, owner):
        if instance is None:
            return self
        if not hasattr(instance, self.private_name):
            setattr(instance, self.private_name, self._get_property(instance))
        return getattr(instance, self.private_name)

    def __set__(self, instance, value):
        prop = self.__get__(instance, instance.__class__)
        prop.clear()
        prop.add(value)

    def _get_property(self, instance) -> RDFProperty:
        return RDFProperty(**self._get_property_kwargs(instance))

    def _get_property_kwargs(self, instance) -> dict[str, Any]:
        return {
            'resource': instance,
            'attr_name': self.name,
            'predicate': self.predicate,
            'required': self.required,
            'repeatable': self.repeatable,
            'values_from': self.values_from,
            'validate': self.validate,
        }


class ObjectProperty(Property):
    def __init__(
            self,
            predicate: URIRef,
            required: bool = False,
            repeatable: bool = False,
            values_from: Container = None,
            validate: Callable[[Any], bool] = None,
            cls: type | str = None,
            embed: bool = False,
    ):
        super().__init__(predicate, required, repeatable, values_from, validate)
        self.object_class = cls
        self.embed = embed

    def _get_property(self, instance) -> RDFObjectProperty:
        if isinstance(self.object_class, str):
            self.object_class = OBJECT_CLASSES[self.object_class]
        return RDFObjectProperty(
            **self._get_property_kwargs(instance),
            object_class=self.object_class,
            embedded=self.embed,
        )

    def __set__(self, instance, value):
        if isinstance(value, URIRef):
            v = value
        elif isinstance(value, EmbeddedObject):
            # here we finally instantiate the embedded object, linking its graph
            # to this instance's graph
            v = value.embed(instance)
        elif hasattr(value, 'uri'):
            v = URIRef(value.uri)
        else:
            # coerce non-URIRef values (e.g., strings) into URIRefs
            v = URIRef(str(value))
        super().__set__(instance, v)


class DataProperty(Property):
    def __init__(
            self,
            predicate: URIRef,
            required: bool = False,
            repeatable: bool = False,
            values_from: Container = None,
            validate: Callable[[Any], bool] = None,
            datatype: URIRef = None,
    ):
        super().__init__(predicate, required, repeatable, values_from, validate)
        self.datatype = datatype

    def _get_property(self, instance) -> RDFDataProperty:
        return RDFDataProperty(
            **self._get_property_kwargs(instance),
            datatype=self.datatype,
        )

    def __set__(self, instance, value):
        # coerce non-Literal values (e.g., strings) into Literals with the datatype
        if not isinstance(value, Literal):
            v = Literal(str(value), datatype=self.datatype)
        else:
            v = value
        super().__set__(instance, v)



from plastron.rdfmapping.resources import RDFResource
from rdflib import URIRef


def test_add_properties():
    resource = RDFResource()
    assert len(resource.rdf_type) == 0
    resource.add_properties(rdf_type=URIRef('http://purl.org/dc/dcmitype/Text'))
    assert len(resource.rdf_type) == 1
    resource.add_properties(rdf_type=URIRef('http://purl.org/dc/dcmitype/Image'))
    assert len(resource.rdf_type) == 2
    assert set(resource.rdf_type.values) == {
        URIRef('http://purl.org/dc/dcmitype/Text'),
        URIRef('http://purl.org/dc/dcmitype/Image'),
    }


def test_add_properties_list():
    resource = RDFResource()
    assert len(resource.rdf_type) == 0
    resource.add_properties(rdf_type=[
        URIRef('http://purl.org/dc/dcmitype/Text'),
        URIRef('http://purl.org/dc/dcmitype/Image'),
    ])
    assert len(resource.rdf_type) == 2
    assert set(resource.rdf_type.values) == {
        URIRef('http://purl.org/dc/dcmitype/Text'),
        URIRef('http://purl.org/dc/dcmitype/Image'),
    }


def test_set_properties():
    resource = RDFResource()
    assert len(resource.rdf_type) == 0
    resource.set_properties(rdf_type=URIRef('http://purl.org/dc/dcmitype/Text'))
    assert len(resource.rdf_type) == 1
    resource.set_properties(rdf_type=URIRef('http://purl.org/dc/dcmitype/Image'))
    assert len(resource.rdf_type) == 1
    assert set(resource.rdf_type.values) == {URIRef('http://purl.org/dc/dcmitype/Image')}


def test_set_properties_list():
    resource = RDFResource()
    assert len(resource.rdf_type) == 0
    resource.set_properties(rdf_type=[
        URIRef('http://purl.org/dc/dcmitype/Text'),
        URIRef('http://purl.org/dc/dcmitype/Image'),
    ])
    assert len(resource.rdf_type) == 2
    assert set(resource.rdf_type.values) == {
        URIRef('http://purl.org/dc/dcmitype/Text'),
        URIRef('http://purl.org/dc/dcmitype/Image'),
    }



from typing import Callable, Any


def default(name: str, *values, required: bool = False):
    """Decorator that set the default values for the named attribute of the decorated
    class. If `required` is `True`, also add a validator that checks that all the default
    values are present in the named attribute."""
    def decorator(cls):
        cls.default_values[name].update(values)
        if required:
            def validator(obj):
                prop = getattr(obj, name)
                return all(v in prop for v in values)
            cls.validators.append(validator)
        return cls
    return decorator


def rdf_type(*values, **kwargs):
    """Shortcut for `@default('rdf_type', *values, **kwargs)`"""
    return default('rdf_type', *values, **kwargs)


def validate(func: Callable[[Any], bool]):
    """Decorator that adds the given function as a validator to the decorated class. The
    validator function is called by `plastron.rdfmapping.resources.RDFResourceBase.is_valid()`,
    with the instance object as its only argument. Its primary use is for validations that
    require checking more than one attribute, as opposed to the per-attribute validators."""
    def decorator(cls):
        cls.validators.append(func)
        return cls
    return decorator



from typing import Callable, Any, Iterable, TypeVar, Type, Iterator, Container

from rdflib import Literal, URIRef
from rdflib.term import Identifier, BNode

from plastron.rdfmapping.embed import EmbeddedObject
from plastron.rdfmapping.validation import ValidationResult, ValidationFailure, ValidationSuccess

T = TypeVar('T')


class RDFProperty:
    """An RDF property"""
    def __init__(
            self,
            resource,
            attr_name: str,
            predicate: URIRef,
            required: bool = False,
            repeatable: bool = False,
            values_from: Container = None,
            validate: Callable[[Any], bool] = None,
    ):
        self.resource = resource
        self.attr_name = attr_name
        self.predicate = predicate
        self.required = required
        self.repeatable = repeatable
        self.values_from = values_from
        self._validate = validate

    @property
    def uri(self) -> URIRef:
        """URI of the predicate"""
        return self.predicate

    @property
    def values(self) -> Iterable:
        """Values of this property"""
        return self.resource.graph.objects(self.resource.uri, self.predicate)

    @property
    def value(self):
        """The first value of this property, or `None` if `values` is empty."""
        try:
            return next(iter(self.values))
        except StopIteration:
            return None

    def __iter__(self):
        return self.values

    def __str__(self):
        return ' '.join(iter(self))

    def __len__(self):
        return len(list(self.values))

    def clear(self):
        """Remove all values from this property."""
        for value in iter(self):
            self.remove(value)

    def add(self, value):
        """Add a single value to this property."""
        self.resource.graph.add((self.resource.uri, self.predicate, value))

    def remove(self, value):
        """Remove a single value from this property."""
        self.resource.graph.remove((self.resource.uri, self.predicate, value))

    def update(self, new_values: Iterable) -> tuple[set, set]:
        """Update this property to have only the new values.

        This method takes the set differences between the current values and the
        new values to construct sets of deleted and inserted values, then removes
        and adds those values, respectively.
        """
        old_values_set = set(self.values)
        new_values_set = set(new_values)
        deleted_values = old_values_set - new_values_set
        inserted_values = new_values_set - old_values_set
        for value in deleted_values:
            self.remove(value)
        for value in inserted_values:
            self.add(value)
        # return the sets so the caller could construct a SPARQL update
        return deleted_values, inserted_values

    def extend(self, values: Iterable):
        """Add each value in the `values` iterable to this property"""
        for value in values:
            self.add(value)

    @property
    def is_valid(self) -> ValidationResult:
        """Checks the validity of this property.

        * If the property is required, there must be at least one value
        * If the property is not repeatable, there must be no more than one value
        * If the property has a `values_from` parameter, then all values must be
          contained by that object (i.e., `value in prop.values_from` must be true)
        * All additional validation functions, such as that given in the `validate`
          parameter of the constructor, must return true values

        If all of these conditions are met, returns a
        `plastron.rdfmapping.validation.ValidationSuccess` object.
        Otherwise, returns a
        `plastron.rdfmapping.validation.ValidationFailure` object.
        """
        if self.required and len(self) == 0:
            return ValidationFailure(self, 'is required')
        if not self.repeatable and len(self) > 1:
            return ValidationFailure(self, 'is not repeatable')
        if self.values_from is not None and any(v not in self.values_from for v in self.values):
            return ValidationFailure(self, f'is not from {self.values_from}')
        if self._validate is not None:
            if all(self._validate(v) for v in self.values):
                return ValidationSuccess(self)
            else:
                return ValidationFailure(self, f'is not {self._validate.__doc__}')
        return ValidationSuccess(self)


class RDFDataProperty(RDFProperty):
    """An RDF property whose values are always RDF literals"""
    def __init__(
            self,
            resource,
            attr_name: str,
            predicate: URIRef,
            required: bool = False,
            repeatable: bool = False,
            values_from: Container = None,
            validate: Callable[[Any], bool] = None,
            datatype: URIRef = None,
    ):
        super().__init__(resource, attr_name, predicate, required, repeatable, values_from, validate)
        self.datatype: URIRef = datatype
        """The datatype of this property"""

    def add(self, value):
        if not isinstance(value, Literal):
            raise TypeError(f'Cannot add a non-Literal value {value} to data property {self.attr_name}')
        super().add(value)

    @property
    def values(self) -> Iterator[Literal]:
        return filter(lambda v: isinstance(v, Literal) and v.datatype == self.datatype, super().values)

    @property
    def languages(self) -> Iterator[str]:
        """Returns an iterator over the language tags of this property's values."""
        return iter(v.language for v in self.values)

    @property
    def is_valid(self) -> ValidationResult:
        """Checks the validity of this property.

        Performs all the validity checks from `RDFProperty.is_valid()`, plus
        requires that all values are RDF literals.

        Even if this property is not marked as repeatable, more than one value
        is allowed so long as each value has a different language tag.
        """
        is_valid_result = super().is_valid
        if not is_valid_result:
            # exception to the superclass rule: if repeatable is False but the only difference
            # in the values is their language, it should be valid
            if not self.repeatable and len(self) > 1:
                if len(set(self.languages)) != len(self):
                    return ValidationFailure(self, 'is not repeatable')
            else:
                return is_valid_result
        # all values must be literals
        if not all(isinstance(v, Literal) for v in self.values):
            return ValidationFailure(self, 'all values must be Literals')
        # if required, all values must be non-blank
        if self.required and any(v.strip() == '' for v in self.values):
            return ValidationFailure(self, 'all values must be non-blank')
        return ValidationSuccess(self)


class RDFObjectProperty(RDFProperty):
    """An RDF property whose values are always URIRefs or RDF blank nodes"""
    def __init__(
            self,
            resource,
            attr_name: str,
            predicate: URIRef,
            required: bool = False,
            repeatable: bool = False,
            values_from: Container = None,
            validate: Callable[[Any], bool] = None,
            object_class: Type[T] = None,
            embedded: bool = False,
    ):
        super().__init__(resource, attr_name, predicate, required, repeatable, values_from, validate)
        self.object_class = object_class
        self.embedded = embedded
        self._object_map = {}

    @property
    def objects(self) -> Iterable[T]:
        """Values of this property, represented as full objects. Requires that this property
        has its `object_class` set; otherwise, it raises an `RDFPropertyError`."""
        if self.object_class is None:
            raise RDFPropertyError(f'No object class defined for the property with predicate {self.predicate}')
        for value in self.values:
            if isinstance(value, Identifier):
                if value not in self._object_map:
                    self._object_map[value] = self.object_class(uri=value, graph=self.resource.graph)
                yield self._object_map[value]
            else:
                yield value

    @property
    def object(self) -> T:
        """The first value of this property, represented as an object, or `None` if `values`
        is empty."""
        try:
            return next(iter(self.objects))
        except StopIteration:
            return None

    def add(self, value):
        if hasattr(value, 'uri'):
            uri = URIRef(value.uri)
            if self.object_class is not None:
                self._object_map[uri] = value
        elif isinstance(value, EmbeddedObject):
            obj = value.embed(self.resource)
            uri = obj.uri
            self._object_map[uri] = obj
        else:
            uri = value
        super().add(uri)

    def remove(self, value):
        if hasattr(value, 'uri'):
            obj = URIRef(value.uri)
            if obj in self._object_map:
                del self._object_map[obj]
        else:
            obj = value
        super().remove(obj)

    @property
    def is_valid(self) -> ValidationResult:
        """Checks the validity of this property.

        Performs all the validity checks from `RDFProperty.is_valid()`, plus
        requires that all values are either URIRefs or RDF blank nodes.
        """
        is_valid_result = super().is_valid
        if not is_valid_result:
            return is_valid_result
        # all values must be URIRefs
        if not all(isinstance(v, URIRef) or isinstance(v, BNode) for v in self.values):
            return ValidationFailure(self, 'all values must be URIs or BNodes')
        return ValidationSuccess(self)


class RDFPropertyError(Exception):
    pass



from rdflib.plugins.sparql import prepareUpdate

from plastron.rdfmapping.graph import TrackChangesGraph


# https://umd-dit.atlassian.net/browse/LIBFCREPO-1739
# rdflib 7.3+ does not currently work with the TrackChangesGraph
# due to an explicit type check in rdflib.plugins.sparql.update.
# A comment above this type check code indicates that the rdflib
# maintainers intend to remove this type check once a series of
# modifications and deprecations in the ConjunctiveGraph and
# Dataset classes. For the time-being, though, we are pinning
# rdflib to a version before this type check was introduced.
#
# This test is in place to ensure that the update functionality
# of the installed version of rdflib works with TrackChangesGraph
def test_rdflib_graph_type_check():
    graph = TrackChangesGraph()
    graph.update(prepareUpdate('DELETE {} INSERT {<> <http://purl.org/dc/terms/title> "Moonpig"} WHERE {}'))
    assert len(graph) == 1



import pytest
from rdflib import URIRef

from plastron.rdfmapping.descriptors import Property, ObjectProperty, DataProperty
from plastron.rdfmapping.properties import RDFProperty, RDFObjectProperty, RDFDataProperty


class MockResource:
    pass


class Foo(MockResource):
    title = Property(URIRef('http://purl.org/dc/terms/title'), required=True, repeatable=False)
    subject = ObjectProperty(URIRef('http://purl.org/dc/terms/subject'), required=True, repeatable=False)
    identifier = DataProperty(URIRef('http://purl.org/dc/terms/identifier'), required=True, repeatable=False)


@pytest.mark.parametrize(
    ('prop_name', 'prop_type'),
    [
        ('title', Property),
        ('subject', ObjectProperty),
        ('identifier', DataProperty),
    ]
)
def test_owner_get(prop_name, prop_type):
    prop = getattr(Foo, prop_name)

    assert isinstance(prop, prop_type)
    assert prop.name == prop_name
    assert prop.required
    assert not prop.repeatable


@pytest.mark.parametrize(
    ('prop_name', 'prop_type'),
    [
        ('title', RDFProperty),
        ('subject', RDFObjectProperty),
        ('identifier', RDFDataProperty),
    ]
)
def test_instance_get(prop_name, prop_type):
    instance = Foo()
    prop = getattr(instance, prop_name)

    assert isinstance(prop, prop_type)
    assert prop.attr_name == prop_name
    assert prop.required
    assert not prop.repeatable



from typing import Optional, ItemsView


class ValidationResult:
    def __init__(self, prop: Optional[object] = None, message: Optional[str] = ''):
        self.prop = prop
        self.message = message

    def __str__(self):
        return self.message

    def __bool__(self):
        raise NotImplementedError


class ValidationFailure(ValidationResult):
    def __bool__(self):
        return False


class ValidationSuccess(ValidationResult):
    def __bool__(self):
        return True


class ValidationResultsDict(dict):
    @property
    def ok(self):
        return len(self.failures()) == 0

    def failures(self) -> ItemsView[str, ValidationFailure]:
        return {k: v for k, v in self.items() if isinstance(v, ValidationFailure)}.items()

    def successes(self) -> ItemsView[str, ValidationSuccess]:
        return {k: v for k, v in self.items() if isinstance(v, ValidationSuccess)}.items()



from typing import TypeVar, Callable, Type, Mapping, Any
from uuid import uuid4

from rdflib import URIRef
from urlobject import URLObject

T = TypeVar('T')


class EmbeddedObject:
    """Wrapper object to delay instantiation of inline-specified objects
    that should be embedded in their parent instance (i.e., share a graph
    object)."""
    def __init__(self, cls: Type[T], fragment_id: str = None, **kwargs):
        self.cls: Type[T] = cls
        """Model class to use for the embedded object"""

        self.fragment_id: str = fragment_id or str(uuid4())
        """Fragment identifier for the embedded object. Defaults to a new UUID."""

        self.kwargs: Mapping[str, Any] = kwargs
        """Keyword arguments to use when constructing the embedded object."""

    def embed(self, instance) -> T:
        """Instantiate an object with the stored class and keyword arguments,
        with a URI from the parent instance object plus the stored fragment ID.
        The `instance` object must have `uri` and `graph` attributes."""
        return self.cls(
            uri=URIRef(URLObject(instance.uri).with_fragment(self.fragment_id)),
            graph=instance.graph,
            **self.kwargs,
        )


def embedded(cls: Type[T]) -> Callable[..., EmbeddedObject]:
    """Function to support an alternative syntax of calling the EmbeddedObject
    constructor. Instead of:

    ```python
    r = MyModel(
        prop=EmbeddedObject(
            MyOtherModel,
            foo=Literal('bar'),
        ),
    )
    ```

    You can use:

    ```python
    r = MyModel(
        prop=embedded(MyOtherModel)(
            foo=Literal('bar'),
        ),
    )
    ```
    """
    def _embedded(**kwargs):
        return EmbeddedObject(cls, **kwargs)
    return _embedded
